/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <stdlib.h>
#include <string.h>

#include "amiga.h"


typedef struct{
    int width;
    int height;
    cairo_surface_t *surface;
}page_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    page_t page[2];
};


static const char **
_wb_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-amiga-icon",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_wb_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->page[0].surface = NULL;
    h->page[1].surface = NULL;
    return h;
}


static void
_wb_free(abydos_plugin_handle_t *h)
{
    int i;
    for(i=0;i<h->info->page_count;++i)
        if(h->page[i].surface)
            cairo_surface_destroy(h->page[i].surface);
    free(h);
}


#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))


static int
_page_from_data(page_t *p, const char *data, const char *end, const amiga_cairo_color_t *palette)
{
    const uint8_t *bitplane[8];
    int depth;
    int stride;
    int i;
    p->width = get16(data+4);
    p->height = get16(data+6);
    depth = get16(data+8);
    stride = amiga_plane_stride(p->width);
    for(i=0;i<depth;++i)
        bitplane[i] = (uint8_t*)data + 20 + stride*p->height*i;
    p->surface = abydos_image_surface_create_from_bitplanes(bitplane, p->width, p->height, depth, palette, 0);
    return 20 + stride*p->height*depth;
}


static int
_wb_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    static const amiga_cairo_color_t palette1[4] = {0x55AAFF, 0xFFFFFF, 0x000000, 0xFF8800};
    static const amiga_cairo_color_t palette2[8] = {0x959595, 0x000000, 0xFFFFFF, 0x3B67A2,
                                                    0x7B7B7B, 0xAFAFAF, 0xAA907C, 0xFFA997};
    const amiga_cairo_color_t *palette;
    //const uint8_t *p = data;
    const char *end;
    int bytes;

    if(len<98)
        return -1;
    if(get16(data) != 0xe310 || data[2] != 0 || data[3] != 1)
        return -1;

    switch(data[45]){
    case 0: /* Old Style icon */
        palette = palette1;
        break;
    case 1: /* NewIcons */
        palette = palette2;
        break;
    default:
        return -1;
    }

    end = data + len;
    data += 78 + (get32(data+66) ? 56 : 0);
    bytes = _page_from_data(&h->page[0], data, end, palette);
    if(bytes < 0)
        return -1;
    data += bytes;
    bytes = _page_from_data(&h->page[1], data, end, palette);
    if(bytes < 0)
        return -1;

    h->info->page_count = 2;
    return 0;
}


static cairo_surface_t *
_wb_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->page[page].surface);
}


static void
_wb_get_page_info(abydos_plugin_handle_t *h, int page, abydos_plugin_page_info_t *info)
{
    info->width = h->page[page].width;
    info->height = h->page[page].height;
    info->pixel_ratio = 0.5;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _wb_get_mime_types,
        .new = _wb_new,
        .free = _wb_free,
        .create_from_data = _wb_create_from_data,
        .get_image_surface = _wb_get_image_surface,
        .get_page_info = _wb_get_page_info
    };
    return &p;
}
