/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <stdlib.h>                     /* malloc() */
#include <string.h>                     /* memcpy() */

#include "vector.h"

#ifndef NULL
# define NULL ((void*)0)
#endif


#define MIN_SIZE 16

array_t *
array_init(array_t *a,int esize)
{
    a->esize = esize;
    if(esize){
        a->allocation = a->esize * MIN_SIZE;
        a->data = malloc(a->allocation);
    }
    a->members = 0;
    return a;
}


void array_done(array_t *a)
{
    if(a->esize)
        free(a->data);
}


void *
array_append(array_t *a, int count){
    int index = a->members;
    int newsize;
    a->members += count;
    newsize = a->members * a->esize;
    if(newsize > a->allocation){
        while(newsize > a->allocation)
            a->allocation *= 2;
        a->data = realloc(a->data, a->allocation);
    }
    return a->data + index*a->esize;
}


void *
array_insert(array_t *a,int pos,int count){
    int move_members = a->members - pos;
    void *p = array_append(a,count);
    if(move_members > 0){
        p = a->data + pos*a->esize;
        memmove(p + count*a->esize, p, move_members*a->esize);
    }
    return p;
}


void
array_delete(array_t *a, int pos, int count){
    int move_members = a->members - (pos+count);
    if(move_members > 0)
        memmove(a->data + pos*a->esize, a->data + (pos+count)*a->esize, move_members*a->esize);
    a->members -= count;
}


void *
array_copy(array_t *a)
{
    size_t size = array_bytes(a);
    void *copy = malloc(size);
    memcpy(copy, a->data, size);
    return copy;
}



vector_buffer_t *
vector_buffer_init(vector_buffer_t *vb, int shape_size, int keyframe_size, int anim_size)
{
    array_init(&vb->point_array,sizeof(point_t));
    array_init(&vb->segment_array,sizeof(vector_segment_t));
    array_init(&vb->shape_array,shape_size);
    array_init(&vb->keyframe_array,keyframe_size);
    array_init(&vb->anim_array,anim_size);
    return vb;
}


void
vector_buffer_done(vector_buffer_t *vb)
{
    array_done(&vb->point_array);
    array_done(&vb->segment_array);
    array_done(&vb->shape_array);
    array_done(&vb->keyframe_array);
    array_done(&vb->anim_array);
}


point_t *
vector_buffer_push_segment(vector_buffer_t *vb, vector_path_t *path, vector_segment_type_t type)
{
    vector_segment_t *seg = vector_buffer_add_segment(vb);
    int points;
    seg->type = type;
    seg->point = vector_buffer_point_count(vb);
    ++path->segment_count;
    switch(type){
    case VECTOR_GOTO:
    case VECTOR_LINE:
        points = 1;
        break;
    case VECTOR_CURVE:
        points = 3;
        break;
    case VECTOR_CLOSE:
        return NULL;
    }
    path->point_count += points;
    return vector_buffer_add_points(vb, points);
}
