/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdlib.h>                     /* malloc(), free() */

#include "avif/avif.h"


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_avif_get_mime_types()
{
    static const char *mime_types[] = {
        "image/avif",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_avif_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    return h;
}


static void
_avif_free(abydos_plugin_handle_t *h)
{
    free(h);
}


static cairo_surface_t *
_surface_from_image(avifImage *image)
{
    //int last_channel = image->depth-1;
    uint8_t *r_src = image->rgbPlanes[AVIF_CHAN_R];
    uint8_t *g_src = image->rgbPlanes[AVIF_CHAN_G];
    uint8_t *b_src = image->rgbPlanes[AVIF_CHAN_B];
    uint8_t *a_src;
    int r_skip = image->rgbRowBytes[AVIF_CHAN_R] - image->width;
    int g_skip = image->rgbRowBytes[AVIF_CHAN_G] - image->width;
    int b_skip = image->rgbRowBytes[AVIF_CHAN_B] - image->width;
    int a_skip;
    cairo_surface_t *surface;
    uint32_t *dst;
    int skip;
    if(image->alphaPlane){
        a_src = image->alphaPlane;
        a_skip = image->alphaRowBytes - image->width;
        surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,image->width,image->height);
    }else{
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,image->width,image->height);
    }
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - image->width;
    if(avifImageUsesU16(image)){
        /* FIXME!! */
    }else{
        int y;
        for(y=0;y<image->height;++y){
            int x;
            for(x=0;x<image->width;++x){
                int r = *r_src++;
                int g = *g_src++;
                int b = *b_src++;
                int a;
                if(image->alphaPlane){
                    a = *a_src++;
                    r = a*r/255;
                    g = a*g/255;
                    b = a*b/255;
                    *dst++ = r<<16|g<<8|b|a<<24;
                }else{
                    *dst++ = r<<16|g<<8|b;
                }
            }
            r_src += r_skip;
            g_src += g_skip;
            b_src += b_skip;
            a_src += a_skip;
            dst += skip;
        }
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_avif_create_from_data(abydos_plugin_handle_t *h,const char *bytes, size_t len)
{
    avifROData d;
    avifDecoder *dec;
    int i;

    d.data = (const uint8_t*)bytes;
    d.size = len;

    dec = avifDecoderCreate();
    if(avifDecoderParse(dec, &d) != AVIF_RESULT_OK)
        goto error;

    h->info->frame_count = dec->imageCount;
    h->frame = (frame_t*)calloc(h->info->frame_count,sizeof(frame_t));
    for(i=0;i<h->info->frame_count;++i){
        if(avifDecoderNextImage(dec) != AVIF_RESULT_OK)
            goto error;
        if(dec->image->width > h->info->width)
            h->info->width = dec->image->width;
        if(dec->image->height > h->info->height)
            h->info->height = dec->image->height;

        avifImageYUVToRGB(dec->image);
        h->frame[i].surface = _surface_from_image(dec->image);
    }
    avifDecoderDestroy(dec);
    return 0;
error:
    avifDecoderDestroy(dec);
    return -1;
}


static cairo_surface_t *
_avif_get_image_surface(abydos_plugin_handle_t *h,
                         int page,
                         cairo_bool_t *layer_visibility,
                         int variant,
                         double time,
                         int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_avif_get_duration(abydos_plugin_handle_t *h,int page,double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _avif_get_mime_types,
        .new = _avif_new,
        .free = _avif_free,
        .create_from_data = _avif_create_from_data,
        .get_image_surface = _avif_get_image_surface,
        .get_duration = _avif_get_duration
    };
    return &p;
}
