/*
 * Copyright (C) 2004 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <endian.h>
#include <string.h>                     /* memset() */

#include "amiga.h"

void
amiga_decode_palette(const uint16_t *amiga_palette, amiga_cairo_color_t *cairo_palette,int colors)
{
    int i;
    for(i=0;i<colors;++i){
        uint16_t col = be16toh(amiga_palette[i]);
        int r = (col>>8)&0x0f;
        int g = (col>>4)&0x0f;
        int b = (col>>0)&0x0f;
#ifdef AMIGA_CAIRO_BITS16
        r = (r<<1) | (r>>3);
        g = (g<<2) | (g>>2);
        b = (b<<1) | (b>>3);
        cairo_palette[i] = (r<<11) | (g<<5) | (b<<0);
#else
        r = (r<<4) | r;
        g = (g<<4) | g;
        b = (b<<4) | b;
        cairo_palette[i] = (r<<16) | (g<<8) | b;
#endif
    }
}


#if 0
void
abydos_copy_line_from_amiga_bitplanes(const uint8_t **amiga_bitplane, amiga_cairo_color_t *dst, int bplcon0, amiga_cairo_color_t *palette)
{
}
#endif


cairo_surface_t *
abydos_image_surface_create_from_bitplanes(const uint8_t **bitplane,
                                           int width, int height, int depth,
                                           const amiga_cairo_color_t *palette, int palette_stepping)
{
    int src_offs = 0;
    int src_stride = amiga_plane_stride(width);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);

    while(height-->0){
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                if(x*8+i >= width)
                    break;
                dst[x*8+i] = palette[b[i]];
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
        palette += palette_stepping;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
abydos_image_surface_create_from_ehb_bitplanes(const uint8_t **bitplane,
                                               int width, int height, int depth,
                                               const amiga_cairo_color_t *palette, int palette_stepping)
{
    int src_offs = 0;
    int src_stride = amiga_plane_stride(width);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int mask = ~(1<<(depth-1));

    while(height-->0){
        uint32_t rgb = 0;
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                int pix;
                if(x*8+i >= width)
                    break;
                pix = b[i];
                if(pix & ~mask){
                    rgb = (palette[pix&mask]>>1) & 0x00777777;
                }else{
                    rgb = palette[pix];
                }
                dst[x*8+i] = rgb;
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
        palette += palette_stepping;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
abydos_image_surface_create_from_ham_bitplanes(const uint8_t **bitplane,

                                               int width, int height, int depth,
                                               const amiga_cairo_color_t *palette, int palette_stepping)
{
    int src_offs = 0;
    int src_stride = amiga_plane_stride(width);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);
    int hold = depth > 6 ? 6 : 4;

    while(height-->0){
        uint32_t rgb = 0;
        int x;
        for(x=0;x<src_stride;++x){
            int plane;
            uint8_t b[8];
            int i;
            memset(b,0,8);
            for(plane=0;plane<depth;++plane){
                for(i=0;i<8;++i)
                    if(bitplane[plane][src_offs+x] & (128>>i))
                        b[i] |= 1<<plane;
            }
            for(i=0;i<8;++i){
                int pix;
                if(x*8+i >= width)
                    break;
                pix = b[i];
                switch(pix>>hold){
                case 0:
                    rgb = palette[pix];
                    break;
                case 1: /* blue */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0xffff00) | pix;
                    break;
                case 2: /* red */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0x00ffff) | pix << 16;
                    break;
                case 3: /* green */
                    pix = pix << (8-hold) & 0xff;
                    pix |= pix >> hold;
                    rgb = (rgb & 0xff00ff) | pix << 8;
                    break;
                }
                dst[x*8+i] = rgb;
            }
        }
        src_offs += src_stride;
        dst += dst_stride;
        palette += palette_stepping;
    }

    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
abydos_image_surface_create_from_amiga_bitplanes(const uint8_t **bitplane,
                                                 int width, int height, int depth, int bplcon0,
                                                 const amiga_cairo_color_t *palette, int palette_size, int palette_stepping)
{
    if(1<<depth > palette_size){
        if(bplcon0 & BPLCON0_HAM){
            DEBUG_p("HAM");
            return abydos_image_surface_create_from_ham_bitplanes(bitplane, width, height, depth, palette, palette_stepping);
        }
        if(!(bplcon0 & BPLCON0_DPF) && 6==depth){
            DEBUG_p("EHB");
            return abydos_image_surface_create_from_ehb_bitplanes(bitplane, width, height, depth, palette, palette_stepping);
        }
    }
    return abydos_image_surface_create_from_bitplanes(bitplane, width, height, depth, palette, palette_stepping);
}


#if 0
cairo_surface_t *
abydos_image_surface_create_from_amiga_pixels(const uint8_t *pixels,
                                              int width, int height,
                                              int stride, int bplcon0,
                                              amiga_cairo_color_t *palette,
                                              int palette_stepping)
{
}
#endif
