/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <byteswap.h>
#include <debug.h>
#include <endian.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc() */
#include <string.h>                     /* memset() */

#include "amiga.h"
#include "cycle.h"
#include "iff.h"

#define ILBM_BMHD IFF_ID('B','M','H','D')
#define ILBM_CMAP IFF_ID('C','M','A','P')
#define ILBM_CAMG IFF_ID('C','A','M','G')
#define ILBM_CRNG IFF_ID('C','R','N','G')
#define ILBM_BODY IFF_ID('B','O','D','Y')

#define ILBM_ANNO IFF_ID('A','N','N','O')
#define ILBM_CTBL IFF_ID('C','T','B','L')
#define ILBM_BEAM IFF_ID('B','E','A','M')
#define ILBM_PCHG IFF_ID('P','C','H','G')
#define ILBM_SHAM IFF_ID('S','H','A','M')
#define ILBM_DPAN IFF_ID('D','P','A','N')
#define ILBM_DPPS IFF_ID('D','P','P','S')

#define ILBM_ANHD IFF_ID('A','N','H','D')
#define ILBM_DLTA IFF_ID('D','L','T','A')

#define ILBM_VDAT IFF_ID('V','D','A','T')

#define DPST_DPAH IFF_ID('D','P','A','H')

#define VDLT_ADAT IFF_ID('A','D','A','T')

#define TEST_ONE_PALETTE
#define CRNG_MAX 16

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif

#define CLAMP_BYTE(a) MAX(0,MIN(255,a))

typedef struct{
    enum{
        IMG_NONE,
        IMG_SURFACE,
        IMG_BITPLANES,
        IMG_PLANAR
    }type;
    union{
        cairo_surface_t *surface;
        uint8_t **bitplane;
        uint8_t *planar;
    };
    ssize_t planar_length;
}image_t;

typedef struct{
    int compression;
    int mask;
    int width;
    int height;
    int x;
    int y;
    int duration;
    uint32_t bits;
}delta_info_t;

typedef struct{
    uint32_t type;
    int width;
    int height;
    int depth;
    double pixel_ratio;
    int screen_width;
    int screen_height;
    int compression;
    int colors;
    amiga_cairo_color_t *palette;
    int use_palette_stepping;

    uint32_t camg;

    color_cycle_t cycle[CRNG_MAX];
    int cycle_count;

    int frame_count;
    double duration_multiplier;
    int default_duration;
    delta_info_t di;
    image_t image;
}frame_info_t;


static image_t *
_image_init(image_t *img)
{
    img->type = IMG_NONE;
    return img;
}


static void
_image_done(image_t *img)
{
    switch(img->type){
    case IMG_NONE:
        break;
    case IMG_SURFACE:
        cairo_surface_destroy(img->surface);
        break;
    case IMG_BITPLANES:
        free(img->bitplane[0]);
        free(img->bitplane);
        break;
    case IMG_PLANAR:
        free(img->planar);
        break;
    }
}


static frame_info_t *
_frame_info_init(frame_info_t *fi)
{
    fi->width = 0;
    fi->height = 0;
    fi->pixel_ratio = 0;
    fi->colors = 0;
    fi->palette = NULL;
    fi->use_palette_stepping = 0;
    fi->camg = 0;
    fi->cycle_count = 0;
    fi->frame_count = -1;
    /* Use jiffies (1/60 sec) as the default. */
    fi->duration_multiplier = 1.0/60;
    fi->default_duration = 0;
    _image_init(&fi->image);
    return fi;
}


static void
_frame_info_done(frame_info_t *fi)
{
    if(fi->palette)
        free(fi->palette);
    _image_done(&fi->image);
}


typedef struct{
    cairo_surface_t *surface;
    double duration;
}frame_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    frame_t *frame;
};


static const char **
_ilbm_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-ilbm",
        "video/x-anim",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_ilbm_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->frame = NULL;
    info->threadsafe = 1;
    return h;
}


static void
_ilbm_free(abydos_plugin_handle_t *h)
{
    if(h->frame){
        int i;
        for(i=0;i<h->info->frame_count;++i)
            if(h->frame[i].surface)
                cairo_surface_destroy(h->frame[i].surface);
        free(h->frame);
    }
    free(h);
}


#define get8(a) (*(uint8_t*)(a))
#define get16(a) be16toh(*(uint16_t*)(a))
#define get32(a) be32toh(*(uint32_t*)(a))

static double
_pixel_ratio_from_screen_size(int width, int height)
{
    /* try to guess aspect ratio from screen size */
    if(height <= 480){
        /* assume this is an 4:3 screen */
        return ((double)height/width) * (4.0/3);
    }else{
        /* higher resolutions are assumed to have square pixels independant
           of their aspect ratio. */
        return 1;
    }
}


static int
_handle_bmhd(frame_info_t *fi, const char *data, ssize_t len)
{
    /*
       0 u16 width
       2 u16 height
       4 u16 xoff
       6 u16 yoff
       8 u8  planes
       9 u8  masking
      10 u8  compression
      11 u8  pad
      12 u16 transparent
      14 u8  x aspect
      15 u8  y aspect
      16 u16 screen width
      18 u16 screen height
    */
    int x_aspect;
    int y_aspect;

    if(len<20)
        return -1;

    fi->width = get16(data+0);
    fi->height = get16(data+2);
    fi->depth = get8(data+8);
    fi->compression = get8(data+10);

    x_aspect = get8(data+14);
    y_aspect = get8(data+15);
    fi->screen_width = get16(data+16);
    fi->screen_height = get16(data+18);
    DEBUG_p("X aspect = %d, Y aspect = %d",x_aspect,y_aspect);
    DEBUG_p("screen width : %d",fi->screen_width);
    DEBUG_p("screen height: %d",fi->screen_height);
    if(x_aspect != y_aspect){
        //fi->pixel_ratio = (double)x_aspect/y_aspect;
        fi->pixel_ratio = (double)x_aspect/y_aspect;
    }else if(fi->screen_width && fi->screen_height){
        fi->pixel_ratio = _pixel_ratio_from_screen_size(fi->screen_width, fi->screen_height);
    }

    return 0;
}


static int
_handle_anhd(delta_info_t *di, const char *data, ssize_t len)
{
    /*
       0 u8  compression
       1 u8  mask
       2 u16 width
       4 u16 height
       6 u16 x
       8 u16 y
      10 u32 abstime
      14 u32 reltime
      18 u8  interleave
      19 u8  pad
      20 u32 bits

      additional 16 bytes padding
    */

    di->compression = get8(data+0);
    di->mask = get8(data+1);
    di->width = get16(data+2);
    di->height = get16(data+4);
    di->x = get16(data+6);
    di->y = get16(data+8);
    di->duration = get32(data+14);
    di->bits = get32(data+20);

    return 0;
}


static int
_handle_cmap(frame_info_t *fi, const char *data, ssize_t len)
{
    int i;
    fi->colors = len/3;
    DEBUG_p("%d colors",fi->colors);
    fi->palette = (amiga_cairo_color_t*)malloc(fi->colors * fi->height * sizeof(amiga_cairo_color_t));
    for(i=0;i<fi->colors;++i)
        fi->palette[i] = get8(data+i*3+0)<<16 | get8(data+i*3+1)<<8 | get8(data+i*3+2);
    return 0;
}


static int
_handle_camg(frame_info_t *fi, const char *data, ssize_t len)
{
    if(len>=4)
        fi->camg = get32(data);
    return 0;
}


static int
_handle_crng(frame_info_t *fi, const char *data, ssize_t len)
{
    int rate;
    int flags;
    int first;
    int last;
    color_cycle_t *cycle;
    if(fi->cycle_count >= CRNG_MAX){
        return 0;
    }
    if(len < 8)
        return -1;
    flags = get16(data+4);
    if(!flags)
        return 0;
    rate = get16(data+2);
    /* this special case also means off */
    if(0==rate || (1==flags && 36==rate))
        return 0;
    DEBUG_p("rate: %d",rate);
    first = (uint8_t)data[6];
    last = (uint8_t)data[7];
    if(first == last)
        return 0;
    cycle = &fi->cycle[fi->cycle_count++];
    cycle->direction = 2==flags ? CCD_RIGHT : CCD_LEFT;
    cycle->duration = 4096.0 / (15*rate);
    cycle->first = first;
    cycle->last = last;
    return 0;
}


static int
_handle_ctbl(frame_info_t *fi, const char *data, ssize_t len)
{
    int entries = len/2;
    int i;

    if(fi->palette){
        entries = MIN(entries,fi->colors * fi->height);
    }else{
        fi->colors = entries/fi->height;
        fi->palette = (amiga_cairo_color_t*)malloc(entries*sizeof(amiga_cairo_color_t));
    }
    fi->use_palette_stepping = 1;

    for(i=0;i<entries;++i){
        int col = get16(data+i*sizeof(uint16_t));
        int r = (col>>8)&0x0f;
        int g = (col>>4)&0x0f;
        int b = (col>>0)&0x0f;
        fi->palette[i] = r<<20 | r<<16 | g<<12 | g<<8 | b<<4 | b;
    }
    return 0;
}


typedef struct{
    const uint8_t *start;
    const uint8_t *tree;
    const uint8_t *data;
    const uint8_t *end;
    int bits;
}huffman_state_t;


static void
_huffman_init(huffman_state_t *state, const uint8_t *data, size_t len, size_t tree_len)
{
    state->start = data;
    state->tree = data+tree_len-2;
    state->data = data+tree_len;
    state->end = data+len;
    state->bits = 0;
}


static int
_huffman_get_bytes(huffman_state_t *state, uint8_t *dst, int bytes)
{
    while(bytes-->0){
        const uint8_t *tree = state->tree;
        for(;;){
            int bit;
            if(!(state->bits&0xff)){
                if(state->data>=state->end)
                    return -1;
                state->bits = (*state->data++)<<1 | 1;
            }
            bit = state->bits&256;
            state->bits<<=1;
            if(bit){
                int hi = tree[0];
                int lo = tree[1];
                if(hi < 0x80){
                    *dst++ = lo;
                    break;
                }
                tree += (hi - 0x100) << 8 | lo;
                if(tree < state->start)
                    return -1;
            }else{
                tree -= 2;
                if(tree < state->start)
                    return -1;
                if((tree[0] & 0x81) == 1){
                    *dst++ = tree[1];
                    break;
                }
            }
        }
    }
    DEBUG_p("%p < %p > %p",state->start,state->data,state->end);
    return 0;
}


static int
_handle_pchg(frame_info_t *fi, const char *data, ssize_t len)
{
    int compression;
    int flags;
    int startline;
    int line_count;
    //int min_reg;
    //int max_reg;
    int tree_size;
    int original_size;
    int line_length;
    const uint8_t *linemask;
    int linemask_len;
    size_t len2;
    uint8_t *allocated_data = NULL;
    const uint8_t *pchg_data;
    int pchg_len;
    huffman_state_t huff;
    int i;

    DEBUG_p("chunk size: %d",(int)len);

    if(len<20)
        return -1;
    compression = get16(data+0);
    flags = get16(data+2);
    startline = get16(data+4);
    line_count = get16(data+6);
    //min_reg = get16(data+8);
    //max_reg = get16(data+10);
    tree_size = get16(data+12);
    original_size = get32(data+16);

    /* FIXME!! make use of startline */

    DEBUG_p("original size: %d",original_size);

    data += 20;
    len -= 20;

    linemask_len = ((line_count+31)>>5)*4;
    DEBUG_p("line mask size: %d",linemask_len);

    switch(compression){
    case 0:
        DEBUG_p("uncompressed");
        pchg_data = (uint8_t*)data;
        pchg_len = len;
        break;
    case 1:
        DEBUG_p("huffman compressed");
        pchg_len = get32(data+4);
        pchg_data = allocated_data = malloc(pchg_len);
        len2 = get32(data);
        DEBUG_p("tree size %d / %d / %d",tree_size,(int)len2,(int)len);
        _huffman_init(&huff, (const uint8_t*)data+8, len-8, len2);
        if(_huffman_get_bytes(&huff, allocated_data, pchg_len)<0)
            goto error;
        DEBUG_p("successfully decompressed %d bytes",(int)pchg_len);
        break;
    default:
        DEBUG_p("unknown compression");
        return -1;
    }

    if(linemask_len > pchg_len)
        goto error;
    linemask = (const uint8_t*)pchg_data;
    pchg_data += linemask_len;
    pchg_len -= linemask_len;

    /* FIXME!! what if there was no CMAP? */

    line_length = fi->colors*sizeof(amiga_cairo_color_t);
    fi->use_palette_stepping = 1;

    DEBUG_p("flags: 0x%x",flags);
    if(!(flags&1)){
        for(i=0;i<line_count;++i){
            int change_count;
            int j;
            if(i > 0)
                memcpy(fi->palette+i*fi->colors,fi->palette+(i-1)*fi->colors,line_length);
            DEBUG_p("mask: 0x%x",linemask[i>>3]);
            if(!(linemask[i>>3] & (128>>(i&31))))
                continue;
            DEBUG_p("there are changes for line %04d",i);
            if(pchg_len < 2)
                goto error;
            change_count = get16(pchg_data);
            len2 = 2 + change_count*6;
            DEBUG_p("changes: %d",change_count);
            if(len2 > pchg_len){
                DEBUG_p("out of data");
                goto error;
            }
            for(j=0;j<change_count;++j){
                int index = get16(pchg_data+2+j*6);
                int r = get8(pchg_data+2+j*6+3);
                int g = get8(pchg_data+2+j*6+4);
                int b = get8(pchg_data+2+j*6+5);
                fi->palette[i*fi->colors+index] = r<<16|g<<8|b;
            }

            pchg_data += len2;
            pchg_len -= len2;
        }
    }else{
        for(i=0;i<line_count;++i){
            int change_count16;
            int change_count32;
            int j;
            if(i > 0)
                memcpy(fi->palette+i*fi->colors,fi->palette+(i-1)*fi->colors,line_length);
            //DEBUG_p("mask: 0x%x",linemask[i>>3]);
            if(!(linemask[i>>3] & (128>>(i&7))))
                continue;
            DEBUG_p("there are changes for line %04d",i);
            if(pchg_len < 2)
                goto error;
            change_count16 = get8(pchg_data);
            change_count32 = get8(pchg_data+1);
            len2 = 2 + change_count16*2 + change_count32*2;
            DEBUG_p("changes16: %d",change_count16);
            DEBUG_p("changes32: %d",change_count32);
            if(len2 > pchg_len){
                DEBUG_p("out of data");
                goto error;
            }
            for(j=0;j<change_count16;++j){
                int change = get16(pchg_data+2+j*2);
                int r = (change>>8) & 0xf;
                int g = (change>>4) & 0xf;
                int b = (change>>0) & 0xf;
                r = r<<4|r;
                g = g<<4|g;
                b = b<<4|b;
                fi->palette[i*fi->colors+(change>>12)] = r<<16|g<<8|b;
            }
            for(j=0;j<change_count32;++j){
                int change = get16(pchg_data+2+change_count16*2+j*2);
                int r = (change>>8) & 0xf;
                int g = (change>>4) & 0xf;
                int b = (change>>0) & 0xf;
                r = r<<4 | r;
                g = g<<4 | g;
                b = b<<4 | b;
                fi->palette[i*fi->colors+16+(change>>12)] = r<<16|g<<8|b;
            }
            pchg_data += len2;
            pchg_len -= len2;
        }
    }
    if(allocated_data)
        free(allocated_data);
    return 0;
error:
    if(allocated_data)
        free(allocated_data);
    return -1;
}


static int
_handle_sham(frame_info_t *fi, const char *data, ssize_t len)
{
    int entries;
    data += 2;
    len -= 2;
    if(fi->height*32 == len)
        return _handle_ctbl(fi, data, len);
    entries = len/2;
    if(entries == 16*(fi->height/2)){
        int y;
        DEBUG_p("colors: %d",fi->colors);

        if(fi->palette){
            entries = MIN(entries,fi->colors * (fi->height/2));
        }else{
            fi->colors = entries/fi->height;
            fi->palette = (amiga_cairo_color_t*)malloc(entries*sizeof(amiga_cairo_color_t));
        }
        fi->use_palette_stepping = 1;

        for(y=0;y<fi->height;++y){
            if(y&1){
                memcpy(&fi->palette[y*fi->colors],&fi->palette[(y-1)*fi->colors],fi->colors*sizeof(amiga_cairo_color_t));
            }else{
                int i;
                for(i=0;i<fi->colors;++i){
                    int col = get16(data + (i + (y/2)*fi->colors) * sizeof(uint16_t));
                    int r = (col>>8)&0x0f;
                    int g = (col>>4)&0x0f;
                    int b = (col>>0)&0x0f;
                    fi->palette[i + y * fi->colors] = r<<20 | r<<16 | g<<12 | g<<8 | b<<4 | b<<0;
                }
            }
        }
        return 0;
    }
    return -1;
}


static uint8_t **
_get_bitplanes(const uint8_t *src, ssize_t len, frame_info_t *fi)
{
    int stride = amiga_plane_stride(fi->width);
    uint8_t **bitplane;
    int y;

    bitplane = (uint8_t**)malloc(fi->depth*sizeof(uint8_t*));
    bitplane[0] = (uint8_t*)malloc(stride*fi->height*fi->depth*sizeof(uint8_t));
    for(y=1;y<fi->depth;++y)
        bitplane[y] = bitplane[y-1] + stride*fi->height;

    if(2 == fi->compression){
        const uint8_t *end = src+len;
        int runlength;
        int bytes = 0;
        int plane;
        for(plane=0;plane<fi->depth;++plane){
            const uint8_t *next_plane;
            const uint16_t *vals;
            int x;
            if(ILBM_VDAT != *(uint32_t*)src)
                goto error;
            DEBUG_p("valid VDAT plane");
            next_plane = src + 8 + get32(src+4);
            vals = (uint16_t*)(src + 8 + get16(src+8));
            src += 10;
            for(x=0;x<stride;x+=2){
                for(y=0;y<fi->height;++y){
                    int bits;
                    if(bytes <= 0){
                        if(src>=end)
                            goto error;
                        bytes = *src++;
                        if(bytes>=128){
                            bytes = 256-bytes;
                            runlength = -1;
                        }else if(0==bytes){
                            if(vals+1 > (uint16_t*)end){
                                DEBUG_p("out of data 1");
                                goto error;
                            }
                            bytes = be16toh(*vals++);
                            runlength = -1;
                        }else if(1==bytes){
                            /* it's a long runlength */
                            if(vals+2 > (uint16_t*)end){
                                DEBUG_p("out of data 2");
                                goto error;
                            }
                            bytes = be16toh(*vals++);
                            runlength = *vals++;
                        }else{
                            /* it's a short runlength */
                            if(vals+1 > (uint16_t*)end){
                                DEBUG_p("out of data 3");
                                goto error;
                            }
                            runlength = *vals++;
                        }
                    }
                    --bytes;
                    if(runlength>=0){
                        bits = runlength;
                    }else{
                        if(src+1 > end)
                            goto error;
                        bits = *vals++;
                    }
                    *(uint16_t*)(&bitplane[plane][x+y*stride]) = bits;
                }
            }
            src = next_plane;
        }
    }else if(fi->compression){
        const uint8_t *end = src+len;
        int runlength;
        int bytes = 0;
        for(y=0;y<fi->height;++y){
            int plane;
            int x;
            for(plane=0;plane<fi->depth;++plane){
                for(x=0;x<stride;++x){
                    int bits;
                    if(bytes-- <= 0){
                        if(src>=end)
                            goto error;
                        bytes = *src++;
                        if(bytes>=128){
                            /* it's a runlength */
                            bytes = 256-bytes;
                            if(src>=end)
                                goto error;
                            runlength = *src++;
                        }else{
                            runlength = -1;
                        }
                    }
                    if(runlength>=0){
                        bits = runlength;
                    }else{
                        if(src>=end)
                            goto error;
                        bits = *src++;
                    }
                    bitplane[plane][x+y*stride] = bits;
                }
            }
        }
    }else{
        /* Just uninterlace the bitplanes */
        for(y=0;y<fi->height;++y){
            int plane;
            for(plane=0;plane<fi->depth;++plane)
                memcpy(bitplane[plane] + y*stride, src + (y*fi->depth + plane)*stride, stride);
        }
    }
    return bitplane;
error:
    free(bitplane[0]);
    free(bitplane);
    return NULL;
}


static uint8_t **
_clone_bitplanes(uint8_t **src, int width, int height, int depth)
{
    int stride = amiga_plane_stride(width);
    uint8_t **dst;
    int y;
    dst = (uint8_t**)malloc(height*sizeof(uint8_t*));
    dst[0] = (uint8_t*)malloc(stride*height*depth);
    for(y=1;y<depth;++y)
        dst[y] = dst[0] + stride*y*height;
    memcpy(dst[0],src[0],stride*height*depth);
    return dst;
}


static int
_apply_delta(uint8_t **bitplane, int planes, int width, int height, const char *data, ssize_t len, int type)
{
    //const char *end = data+len;
    int stride;
    int plane;
    switch(type){
    case 5:
        if(len<planes*4){
            DEBUG_p("delta chunk too small");
            return -1;
        }
        stride = amiga_plane_stride(width);
        for(plane=0;plane<planes;++plane){
            off_t offs = get32(data+plane*4);
            int x;
            for(x=0;x<stride;++x){
                int y;
                if(offs>=len){
                    DEBUG_p("trying to get operation count past en of chunk");
                    return -1;
                }
                int count = get8(data+offs++);
                y = 0;
                while(count-->0){
                    int op;
                    if(offs>=len){
                        DEBUG_p("trying to get operation past end of chunk");
                        return -1;
                    }
                    op = get8(data+offs++);

                    if(op & 128){
                        int bytes = op & ~128;
                        while(bytes-->0){
                            if(y >= height){
                                DEBUG_p("trying to apply operation past image height");
                                return -1;
                            }
                            if(offs >= len){
                                DEBUG_p("trying to get pixels past end of chunk (op=%d)",op);
                                return -1;
                            }
                            bitplane[plane][x+y*stride] = get8(data+offs++);
                            ++y;
                        }
                    }else if(0==op){
                        int bytes;
                        int val;
                        if(offs+1 >= len){
                            DEBUG_p("trying to get pixels past end of chunk (op=%d)",op);
                            return -1;
                        }
                        bytes = get8(data+offs++);
                        val = get8(data+offs++);
                        while(bytes-->0){
                            if(y >= height){
                                DEBUG_p("trying to apply operation past image height");
                                return -1;
                            }
                            bitplane[plane][x+y*stride] = val;
                            ++y;
                        }
                    }else{
                        y += op;
                    }
                }
            }
        }
    }
    return 0;
}


static int
_get_hame_byte(uint8_t **bitplane, int offs){
    int himask = 128>>((offs&3)<<1);
    int lomask = 64>>((offs&3)<<1);
    int plane;
    int c = 0;
    for(plane=0;plane<4;++plane){
        int bits = bitplane[plane][offs>>2];
        if(bits & himask)
            c |= 16<<plane;
        if(bits & lomask)
            c |= 1<<plane;
    }
    return c;
}


static int
_is_hame(uint8_t **bitplane, int offs){
    int i;
    for(i=0;i<7;++i){
        static const uint8_t magic[7] = {0xa2, 0xf5, 0x84, 0xdc, 0x6d, 0xb0, 0x7f};
        if(_get_hame_byte(bitplane, offs+i) != magic[i])
            return 0;
    }
    switch(_get_hame_byte(bitplane, offs+7)){
    case 0x14: // REG
    case 0x18: // HAM-E
        return 1;
    default:
        return 0;
    }
}


static cairo_surface_t *
_surface_from_hame(uint8_t **bitplane, int width, int height, uint32_t bplcon0)
{
    int lace = (bplcon0 & BPLCON0_LACE);
    int src_offs = 0;
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    uint32_t palette[512];
    int palette_length[2] = {0,0};
    int hame = 0;
    int y;

    memset(palette,0,sizeof(palette));

    for(y=0;y<height;++y){
        int palette_offs = lace && y&1 ? 0 : 256;
        if(_is_hame(bitplane,src_offs)){
            int c;
            palette_offs += palette_length[palette_offs >> 8];
            DEBUG_p("%04d: palette line",y);
            for(c=0;c<64;++c){
                palette[palette_offs + c] = _get_hame_byte(bitplane, src_offs + 8+c*3 + 0) << 16 |
                                            _get_hame_byte(bitplane, src_offs + 8+c*3 + 1) << 8 |
                                            _get_hame_byte(bitplane, src_offs + 8+c*3 + 2);
            }
            palette_length[palette_offs >> 8] = (palette_length[palette_offs >> 8] + 64) & 0xff;
            hame = _get_hame_byte(bitplane,src_offs+7) == 0x18;
            memset(dst,0,width*sizeof(uint32_t)); /* blank the special line */
            src_offs += width;
            dst += width;
        }else{
            int palette_bank = 0;
            uint32_t rgb = 0;
            int x;
            if(hame){
                DEBUG_p("%04d: HAM-E line",y);
                for(x=0;x<width;++x){
                    int c = _get_hame_byte(bitplane,src_offs++);
                    switch(c>>6){
                    case 2: /* red */
                        rgb = (c & 0x3f) << 18 | (rgb & 0x00ffff);
                        break;
                    case 3: /* green */
                        rgb = (c & 0x3f) << 10 | (rgb & 0xff00ff);
                        break;
                    case 1: /* blue */
                        rgb = (c & 0x3f) << 2 | (rgb & 0xffff00);
                        break;
                    default:
                        if (c < 60)
                            rgb = palette[palette_offs + palette_bank + c];
                        else
                            palette_bank = (c - 60) << 6;
                        break;
                    }
                    *dst++ = rgb;
                }
            }else{
                DEBUG_p("%04d: normal line line",y);
                for(x=0;x<width;++x)
                    *dst++ = palette[palette_offs + _get_hame_byte(bitplane,src_offs++)];
            }
        }
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
_get_dctv_value(uint8_t **bitplane, int depth, int offs, int x, amiga_cairo_color_t *palette){
    amiga_cairo_color_t rgb;
    int mask;
    int plane;
    int c = 0;
    offs += x>>3;
    mask = 128>>(x&7);
    for(plane=0;plane<depth;++plane)
        if(bitplane[plane][offs]&mask)
            c |= 1<<plane;
    rgb = palette[c];
    /* The value is 0I0R0B0G.
     * Digital R, G, B are the highest bits of the components.
     * I is the lowest bit of the 4-bit Blue component.
     */
    return (rgb << 2 & 0x40) | (rgb >> 19 & 0x10) | (rgb >> 5 & 4) | (rgb >> 15 & 1);
}


static int
_is_dctv(uint8_t **bitplane, int depth, int offs, amiga_cairo_color_t *palette){
    int r;
    int x;
    if(_get_dctv_value(bitplane, depth, offs, 0, palette) >> 6 != 0)
        return 0;
    r = 0x7d;
    for(x=1;x<256;++x){
        if(_get_dctv_value(bitplane, depth, offs, x, palette) >> 6 == (r & 1))
            return 0;
        if ((r & 1) != 0)
        r ^= 0xc3 << 1;
        r >>= 1;
    }
    return 1;
}


static int
_is_dctv2(uint8_t **bitplane, int width, int depth, uint32_t bplcon0, amiga_cairo_color_t *palette){
    if(!_is_dctv(bitplane, depth, 0, palette))
        return 0;
    if(bplcon0 & BPLCON0_LACE)
        if(!_is_dctv(bitplane, depth, amiga_plane_stride(width), palette))
            return 0;
    return 1;
}


static cairo_surface_t *
_surface_from_dctv(uint8_t **bitplane, int width, int height, int depth, uint32_t bplcon0, amiga_cairo_color_t *palette)
{
    int lace;
    int plane_stride;
    int src_offs;
    cairo_surface_t *surface;
    uint32_t *dst;
    int dst_skip;
    int chroma[2560];
    int y;

    lace = (bplcon0 & BPLCON0_LACE) != 0;
    plane_stride = amiga_plane_stride(width);
    surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    {
        int skiprows = lace ? 2 : 1;
        memset(dst, 0, skiprows*(width+dst_skip)*sizeof(uint32_t));
        src_offs = skiprows * plane_stride;
        dst += skiprows * (width+dst_skip);
        height -= skiprows;
    }

    for(y = 0; y < height; ++y){
        int odd = y >> lace & 1;
        int rgb = 0;
        int o = 0;
        int p = 0;
        int x;
        for(x=0;x < width;++x){
            if ((x & 1) == odd){
                int n = x + 1 < width ? _get_dctv_value(bitplane, depth, src_offs, x, palette) << 1 | _get_dctv_value(bitplane, depth, src_offs, x+1, palette) : 0;
                int i = (o + n) >> 1;
                i = i <= 64 ? 0 : i >= 224 ? 255 : (i - 64) * 8 / 5;

                int u = n + p - (o << 1);
                if (u < 0)
                    u += 3;
                u >>= 2;
                if (((x + 1) & 2) == 0)
                    u = -u;
                int chroma_offs = (x & ~1) | (y & lace);
                int v = y > lace ? chroma[chroma_offs] : 0;
                chroma[chroma_offs] = u;
                if (odd == 0) {
                    u = v;
                    v = chroma[chroma_offs];
                }

                p = o;
                o = n;

                int r = i + v * 4655 / 2560;
                int b = i + u * 8286 / 2560;
                int g = i - (v * 2372 + u * 1616) / 2560;
                rgb = CLAMP_BYTE(r) << 16 | CLAMP_BYTE(g) << 8 | CLAMP_BYTE(b);
            }
            *dst++ = rgb;
        }
        src_offs += plane_stride;
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_planar(const uint8_t *src, ssize_t len, int compression, int width, int height, amiga_cairo_color_t *palette, int colors)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width,height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;

    if(compression){
        const uint8_t *end = src+len;
        int src_stride = width + (width&1);
        int runlength;
        int bytes = 0;
        while(height-->0){
            int x;
            for(x=0;x<src_stride;++x){
                uint32_t pix;
                if(bytes-- <= 0){
                    if(src>=end)
                        goto error;
                    bytes = *src++;
                    if(bytes>=128){
                        /* it's a runlength */
                        bytes = 256-bytes;
                        if(src>=end)
                            goto error;
                        runlength = *src++ % colors;
                    }else{
                        runlength = -1;
                    }
                }
                if(runlength>=0){
                    pix = palette[runlength];
                }else{
                    if(src>=end)
                        goto error;
                    pix = palette[*src++ % colors];
                }
                if(x < width)
                    *dst++ = pix;
            }
            dst += dst_skip;
        }
    }else{
        int src_skip = width&1;
        if((width+src_skip) * height > len)
            goto error;
        while(height-->0){
            int x;
            for(x=0;x<width;++x)
                *dst++ = palette[*src++ % colors];
            src += src_skip;
            dst += dst_skip;
        }
    }
    cairo_surface_mark_dirty(surface);
    return surface;
error:
    cairo_surface_destroy(surface);
    return NULL;
}


static cairo_surface_t *
_surface_from_interlaced_rgb(const uint8_t *src, ssize_t len, frame_info_t *fi)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,fi->width,fi->height);
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_stride = cairo_image_surface_get_stride(surface)/sizeof(uint32_t);

    if(fi->compression){
        const uint8_t *end = src+len;
        int src_stride = amiga_plane_stride(fi->width-1);
        int runlength;
        int bytes = 0;
        int y;
        for(y=0;y<fi->height;++y){
            int plane;
            for(plane=0;plane<fi->depth;++plane){
                /* convert plane bit order BGR to RGB */
                uint32_t mask = bswap_32(1<<plane)>>8;
                int x;
                for(x=0;x<src_stride;++x){
                    int bits;
                    int b;
                    if(bytes-- <= 0){
                        if(src>=end)
                            goto error;
                        bytes = *src++;
                        if(bytes>=128){
                            /* it's a runlength */
                            bytes = 256-bytes;
                            if(src>=end)
                                goto error;
                            runlength = *src++;
                        }else{
                            runlength = -1;
                        }
                    }
                    if(runlength>=0){
                        bits = runlength;
                    }else{
                        if(src>=end)
                            goto error;
                        bits = *src++;
                    }
                    for(b=0;b<8;++b){
                        if(x*8+b >= fi->width)
                            break;
                        if(bits & 1<<(7-b))
                            *dst |= mask;
                        ++dst;
                    }
                }
                dst -= fi->width;
            }
            dst += dst_stride;
        }
    }
    cairo_surface_mark_dirty(surface);
    return surface;
error:
    cairo_surface_destroy(surface);
    return NULL;
}


static cairo_surface_t *
_surface_from_image(image_t *img, int compression, int width, int height, int depth, int camg, amiga_cairo_color_t *palette, int colors, int palette_stepping)
{
    switch(img->type){
    case IMG_NONE:
        return NULL;
    case IMG_SURFACE:
        return cairo_surface_reference(img->surface);
    case IMG_BITPLANES:
        if(width >= 256 && height >= 3 && _is_dctv2(img->bitplane, width, depth, camg, palette)){
            return _surface_from_dctv(img->bitplane, width, height, depth, camg, palette);
        }else{
            return abydos_image_surface_create_from_amiga_bitplanes((const uint8_t**)img->bitplane,width,height,depth,camg,palette,colors,palette_stepping);
        }
    case IMG_PLANAR:
        /* palette stepping not supported (?) */
        return _surface_from_planar(img->planar, img->planar_length, compression, width, height, palette, colors);
    }
    /* not reached */
    return NULL;
}


static cairo_surface_t *
_surface_from_frame_info(frame_info_t *fi, amiga_cairo_color_t *palette)
{
    int palette_stepping;
    if(!palette)
        palette = fi->palette;
    palette_stepping = fi->use_palette_stepping ? fi->colors : 0;
    return _surface_from_image(&fi->image, fi->compression, fi->width, fi->height, fi->depth, fi->camg, palette, fi->colors, palette_stepping);
}


static void
_sanity_check(frame_info_t *fi){
    DEBUG_p("width      : %d",fi->width);
    DEBUG_p("height     : %d",fi->height);
    DEBUG_p("depth      : %d/%d",fi->depth,(fi->camg>>12) & 7);
    DEBUG_p("camg       : %x",fi->camg);
    DEBUG_p("LACE       : %s",fi->camg & BPLCON0_LACE ? "yes" : "no");
    DEBUG_p("UHRES      : %s",fi->camg & BPLCON0_UHRES ? "yes" : "no");
    DEBUG_p("HAM        : %s",fi->camg & BPLCON0_HAM ? "yes" : "no");
    DEBUG_p("HIRES      : %s",fi->camg & BPLCON0_HIRES ? "yes" : "no");
    DEBUG_p("pixel ratio: %f",fi->pixel_ratio);
    DEBUG_p("colors     : %d",fi->colors);

    /*
        When EHB should not be forced:

            width      : 640
            height     : 480
            depth      : 4/0
            camg       : 8004
            LACE       : yes
            UHRES      : no
            HAM        : no
            HIRES      : yes
            pixel ratio: 0.909091
            colors     : 16

        When EHB should be forced:

            width      : 320
            height     : 512
            depth      : 6/4
            camg       : 4084
            LACE       : yes
            UHRES      : yes
            HAM        : no
            HIRES      : no
            pixel ratio: 1.818182
            colors     : 64

    */
    if(fi->camg & BPLCON0_UHRES && fi->colors > 32 && fi->colors >= 1<<fi->depth){
        fi->colors /= 2;
    }

    if(fi->screen_width && fi->screen_height){
        double physical_ratio = fi->pixel_ratio*fi->screen_width/fi->screen_height;
        DEBUG_p("claimed physical screen ratio: %f",physical_ratio);
        if(physical_ratio < 1 || physical_ratio > 2.35){
            /* unreasonable screen ratio */
            fi->pixel_ratio = _pixel_ratio_from_screen_size(fi->screen_width, fi->screen_height);
        }
    }
}


static int
_handle_ilbm(frame_info_t *fi, const char *data, ssize_t len)
{
    while(len >= 8){
        ssize_t chunk_len = get32(data+4);
        if(chunk_len > len-8)
            return -1;
        switch(*(uint32_t*)data){
        case ILBM_BMHD:
            if(_handle_bmhd(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_ANHD:
            if(_handle_anhd(&fi->di,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_ANNO:
            DEBUG_p("ANNO");
            break;
        case ILBM_CMAP:
            if(_handle_cmap(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_CAMG:
            if(_handle_camg(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_DPAN: /* Deluxe Paint Animation */
            DEBUG_p("DPAN size: %d",(int)chunk_len);
            if(chunk_len>=8){
                DEBUG_p("DPAN version: %d", get16(data+8+0));
                DEBUG_p("DPAN frames : %d", get16(data+8+2));
                DEBUG_p("DPAN FPS    : %d", get8(data+8+4));
                /* Set duration multiplier based on FPS. */
                fi->duration_multiplier = 1.0/get8(data+8+4);
            }
            break;
        case ILBM_DPPS: /* Deluxe Paint Preview Storage or Deluxe Paint III Perspective? */
            DEBUG_p("DPPS");
            /* FIXME!! some bits might be related to animation speed */
            if(chunk_len>=108){
                DEBUG_p("DPPS offset  74: %d", get16(data+8+74));
                DEBUG_p("DPPS offset  90: %d", get16(data+8+90));
                DEBUG_p("DPPS offset 106: %d", get16(data+8+106));
            }
            break;
        case ILBM_CRNG:
            if(_handle_crng(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_CTBL:
        case ILBM_BEAM:
            if(_handle_ctbl(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_DLTA:
            if(_apply_delta(fi->image.bitplane,fi->depth,fi->width,fi->height,data+8,chunk_len,fi->di.compression)<-1)
                return -1;
            //frame->surface = abydos_image_surface_create_from_amiga_bitplanes((const uint8_t**)fi->bitplane,fi->width,fi->height,fi->depth,fi->camg,fi->palette,fi->colors,0);
            break;
        case ILBM_PCHG:
            if(_handle_pchg(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_SHAM:
            if(_handle_sham(fi,data+8,chunk_len)<0)
                return -1;
            break;
        case ILBM_BODY:
            if(fi->width <= 0 || fi->height <= 0)
                return -1;
            switch(fi->type){
            case IFF_ILBM:
                if(fi->colors <= 0){
                    fi->image.surface = _surface_from_interlaced_rgb((uint8_t*)data+8,chunk_len,fi);
                    if(fi->image.surface){
                        fi->image.type = IMG_SURFACE;
                    }else{
                        return -1;
                    }
                }else{
                    _sanity_check(fi);
                    fi->image.bitplane = _get_bitplanes((uint8_t*)data+8,chunk_len,fi);
                    if(fi->image.bitplane){
                        fi->image.type = IMG_BITPLANES;
                        if(4==fi->depth && _is_hame(fi->image.bitplane, 0)){
                            cairo_surface_t *surface;
                            fi->width /= 2;
                            fi->pixel_ratio *= 2;
                            surface = _surface_from_hame(fi->image.bitplane,fi->width,fi->height,fi->camg);
                            _image_done(&fi->image);
                            _image_init(&fi->image);
                            if(surface){
                                fi->image.type = IMG_SURFACE;
                                fi->image.surface = surface;
                            }else{
                                return -1;
                            }
                        }
                    }else{
                        return -1;
                    }
                }
                break;
            case IFF_PBM:
                if(fi->colors <= 0)
                    return -1;
                /* FIXME!! copy the planar pixmap instead if we might want
                   to utilize palette cycling */
                fi->image.surface = _surface_from_planar((uint8_t*)data+8,chunk_len,fi->compression,fi->width,fi->height,fi->palette,fi->colors);
                if(fi->image.surface){
                    fi->image.type = IMG_SURFACE;
                }else{
                    return -1;
                }
                break;
            }
            /* nothing else should appear after the BODY */
            return 0;
        default:
            DEBUG_p("unhandled chunk (%4s)",data);
            break;
        }
        chunk_len += chunk_len&1;
        data += chunk_len+8;
        len -= chunk_len+8;
    }
    return 0;
}


static int
_handle_adat(uint8_t *bitplane, frame_info_t *fi, const signed char *data, ssize_t len)
{
    const uint16_t *src;
    const uint16_t *src_end;
    uint16_t *dst;
    const uint16_t *dst_end;
    int stride;

    if(len<2)
        return 0;

    src = (uint16_t*)(data+get16(data));
    src_end = (uint16_t*)(data+len);
    dst = (uint16_t*)bitplane;
    stride = fi->width / 16;
    dst_end = dst + fi->height*stride;

    data += 2;
    while(src < src_end){
        int b = *data++;
        if(0==b){
            int c = (int16_t)be16toh(*src++);
            //DEBUG_p("move %d bytes",c);
            dst += c/2;
            continue;
        }

        if(dst < (uint16_t*)bitplane){
            DEBUG_p("bitplane underrun");
            return -1;
        }else if(dst >= dst_end){
            DEBUG_p("bitplane overrun");
            return -1;
        }

        if(b<0){
            b = -1 -b;
            //DEBUG_p("copy %d words",b);
            if(src + b > src_end)
                return -1;
            while(b-->0){
                *dst ^= *src++;
                dst += stride;
            }
        }else{
            int bits = *src++;
            //DEBUG_p("repeat %d words",b);
            while(b-->0){
                *dst ^= bits;
                dst += stride;
            }
        }
    }
    return 0;
}


static int
_handle_vdlt(frame_info_t *fi, const char *data, ssize_t len)
{
    int plane = 0;
    while(len >= 8 && plane < fi->depth){
        ssize_t chunk_len = get32(data+4);
        if(chunk_len > len-8)
            return -1;
        switch(*(uint32_t*)data){
        case VDLT_ADAT:
            if(_handle_adat(fi->image.bitplane[plane],fi,data+8,chunk_len)<0)
                return -1;
            ++plane;
            break;
        }
        chunk_len += chunk_len&1;
        data += chunk_len+8;
        len -= chunk_len+8;
    }
    return 0;
}


static int
_count_frames(const char *data, ssize_t len)
{
    int count = 0;
    while(len >= 8){
        uint32_t chunk_type = *(uint32_t*)data;
        ssize_t chunk_len = get32(data+4);
        data += 8;
        if(IFF_FORM == chunk_type){
            switch(*(uint32_t*)data){
            case IFF_ILBM:
            case IFF_PBM:
            case IFF_VDLT:
                ++count;
                break;
            }
        }
        data += chunk_len;
        len -= chunk_len+8;
    }
    return count;
}


static int
_handle_dpah(frame_info_t *fi, const char *data, ssize_t len)
{
    if(len>=2){
        fi->default_duration = get16(data+0);
        DEBUG_p("default duration: %d",fi->default_duration);
    }
    if(len>=4)
        fi->frame_count = get16(data+2);
    return 0;
}


static int
_handle_anim(abydos_plugin_handle_t *h, const char *data, ssize_t len, int buffers)
{
    frame_info_t fi;
    uint8_t **buffer[4];
    int res;
    int actual_frame;
    int i;

    h->info->frame_count = _count_frames(data,len);
    if(h->info->frame_count<=0)
        return -1;
    DEBUG_p("found %d frames",h->info->frame_count);
    h->frame = (frame_t*)malloc(h->info->frame_count*sizeof(frame_t));
    for(i=0;i<h->info->frame_count;++i)
        h->frame[i].surface = NULL;
    _frame_info_init(&fi);
    actual_frame = 0;
    for(i=0;i<h->info->frame_count;){
        uint32_t chunk_type = *(uint32_t*)data;
        ssize_t chunk_len = get32(data+4);
        data += 8;
        switch(chunk_type){
        case DPST_DPAH:
            if(_handle_dpah(&fi,data,len)<0)
                return -1;
            break;
        case IFF_FORM:
            fi.type = *(uint32_t*)data;
            switch(fi.type){
            case IFF_ILBM:
            case IFF_VDLT:
                if(i>0)
                    fi.image.bitplane = buffer[i%buffers];
                fi.di.duration = fi.default_duration;
                if(IFF_ILBM == fi.type){
                    res = _handle_ilbm(&fi,data+4,chunk_len-4);
                }else{
                    res = _handle_vdlt(&fi,data+4,chunk_len-4);
                }
                if(res<0){
                    _frame_info_done(&fi);
                    return -1;
                }
                if(0==i){
                    int j;
                    buffer[0] = fi.image.bitplane;
                    for(j=1;j<buffers;++j)
                        buffer[j] = _clone_bitplanes(fi.image.bitplane,fi.width,fi.height,fi.depth);
                }
                /* Discard frames with zero duration. */
                if(fi.di.duration<=0){
                    --h->info->frame_count;
                }else{
                    h->frame[actual_frame].surface = _surface_from_frame_info(&fi, NULL);
                    h->frame[actual_frame].duration = fi.duration_multiplier * fi.di.duration;
                    DEBUG_p("frame %d duration: %f (%f*%d)",i,h->frame[actual_frame].duration,fi.duration_multiplier,fi.di.duration);
                    ++actual_frame;
                }
                ++i;
                break;
            }
            if(fi.width > h->info->width)
                h->info->width = fi.width;
            if(fi.height > h->info->height)
                h->info->height = fi.height;
            break;
        }
        data += chunk_len;
        len -= chunk_len+8;
    }
    _frame_info_done(&fi);
    return 0;
}


static int
_ilbm_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    /* for single frame images */
    frame_info_t fi;
    int ret;
    size_t form_len;
    if(len < 12)
        return -1;
    if(*(uint32_t*)data != IFF_FORM)
        return -1;
    form_len = get32(data+4)-4;
    if(form_len > len-12){
        DEBUG_p("truncated data");
        return -1;
    }
    fi.type = *(uint32_t*)(data+8);
    data += 12;
    switch(fi.type){
    case IFF_ILBM:
    case IFF_PBM:
        _frame_info_init(&fi);
        ret = _handle_ilbm(&fi, data, form_len);
        if(0==ret){
            if(fi.cycle_count>0){
                int i;
                DEBUG_p("got %d color cycles",fi.cycle_count);
                for(i=0;i<fi.cycle_count;++i){
                    color_cycle_t *c = &fi.cycle[i];
                    DEBUG_p("%d: duration: %f",i,c->duration);
                    DEBUG_p("%d: range %d - %d",i,c->first,c->last);
                }
                h->info->frame_count = color_cycle_total_frames(fi.cycle,fi.cycle_count);
                h->frame = (frame_t*)malloc(h->info->frame_count * sizeof(frame_t));
                for(i=0;i<h->info->frame_count;++i){
                    /* cycle the original palette since we create a surface
                       for every frame then ditch the palette (for now). */
                    h->frame[i].surface = _surface_from_frame_info(&fi, fi.palette);
                    h->frame[i].duration = fi.cycle[0].duration;
                    color_cycle_apply_step(&fi.cycle[0], fi.palette, sizeof(amiga_cairo_color_t));
                }
            }else{
                h->frame = (frame_t*)malloc(sizeof(frame_t));
                h->frame->surface = _surface_from_frame_info(&fi,NULL);
            }
        }
        _frame_info_done(&fi);
        if(ret<0)
            return -1;
        h->info->width = fi.width;
        h->info->height = fi.height;
        /* FIXME!! determine if the pixel ratio is reasonable */
        h->info->pixel_ratio = fi.pixel_ratio;
        return 0;
    case IFF_ANIM:
        return _handle_anim(h, data, form_len, 2);
    case IFF_DPST:
        return _handle_anim(h, data, form_len, 1);
    default:
        return -1;
    }
}


static cairo_surface_t *
_ilbm_get_image_surface(abydos_plugin_handle_t *h, int page, cairo_bool_t *layer_visibility, int variant, double time, int frame)
{
    return cairo_surface_reference(h->frame[frame].surface);
}


static void
_ilbm_get_duration(abydos_plugin_handle_t *h, int page, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->frame[i].duration;
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _ilbm_get_mime_types,
        .new = _ilbm_new,
        .free = _ilbm_free,
        .create_from_data = _ilbm_create_from_data,
        .get_image_surface = _ilbm_get_image_surface,
        .get_duration = _ilbm_get_duration
    };
    return &p;
}
