/*
 * Copyright (C) 1999         The Free Software Foundation
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
/*
 * Based on io-ras.c from gdk-pixbuf 2.30
 * Authors: Arjan van de Ven <arjan@fenrus.demon.nl>
 *          Federico Mena-Quintero <federico@gimp.org>
 */
/*
 * Known bugs:
 * Compressed rasterfiles don't work yet
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <endian.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>

#include "buffer.h"

#define HEADER_SIZE 32

/* Sun supported ras_types */
#define RT_OLD          0               /* Raw pixrect image in 68000 byte order */
#define RT_STANDARD     1               /* Raw pixrect image in 68000 byte order */
#define RT_BYTE_ENCODED 2               /* Run-length compression of bytes */
#define RT_FORMAT_RGB   3               /* XRGB or RGB instead of XBGR or BGR */
#define RT_FORMAT_TIFF  4               /* tiff <-> standard rasterfile */
#define RT_FORMAT_IFF   5               /* iff (TAAC format) <-> standard rasterfile */
#define RT_EXPERIMENTAL 0xffff          /* Reserved for testing */

/* Sun supported ras_maptypes */
#define RMT_NONE        0               /* ras_maplength is expected to be 0 */
#define RMT_EQUAL_RGB   1               /* red[ras_maplength/3],green[],blue[] */
/* Sun registered ras_maptypes */
#define RMT_RAW         2

/*
 * Some software, such as Imagemagick, uses the 8 bit padding if the
 * 32-format as an alpha channel. This is not originally part of the format
 * and therefore disabled by default.
 */
/* #define SUPPORT_ALPHA */

/*
   Header structure for sunras files.
   All values are in big-endian order on disk

   Note: Every scanline is padded to be a multiple of 16 bits
 */
typedef struct{
    uint32_t magic;
    uint32_t width;
    uint32_t height;
    uint32_t depth;
    uint32_t length;
    uint32_t type;
    uint32_t maptype;
    uint32_t maplength;
}header_t;

typedef void(*copy_pixels_func_t)(uint32_t*,const uint8_t*,size_t,const uint32_t*);

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    struct{
        abydos_plugin_info_func_t info_func;
        abydos_plugin_progress_func_t progress_func;
        void *userdata;

        int cmap_size;

        int line_bytes;                 /* The width of a line in bytes */
        uint8_t *line_buf;              /* Buffer for 1 line */
        int LineDone;                   /* # of bytes in LineBuf */
        int Lines;                      /* # of finished lines */

        int DecoderState;

        int header_type;
        int header_maptype;

        copy_pixels_func_t copy_pixels;
    }state;
    enum{
        PS_HEADER,
        PS_CMAP,
        PS_DATA,
        PS_IGNORE
    }parse_state;
    abydos_buffer_t buf;
    uint32_t *cmap;

    cairo_surface_t *surface;
};


static const char **
_sunras_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-cmu-raster",
        "image/x-sun-raster",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_sunras_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_sunras_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static void
_sunras_progressive_begin(abydos_plugin_handle_t *h,abydos_plugin_info_func_t info_func,abydos_plugin_progress_func_t progress_func,void *userdata)
{
    h->state.info_func = info_func;
    h->state.progress_func = progress_func;
    h->state.userdata = userdata;

    h->state.line_buf = NULL;
    h->state.LineDone = 0;
    h->state.Lines = 0;

    h->state.DecoderState = 0;

    h->parse_state = PS_HEADER;
    abydos_buffer_clear(&h->buf);
    h->cmap = NULL;
}


static void
_copy_pixels_bgrx(uint32_t *dst, const uint8_t *src, size_t len,const uint32_t *cmap)
{
#ifdef SUPPORT_ALPHA
    int has_alpha = 0;
    int i;
    for(i=0;i<len;++i)
        if(src[i*4+3]){
            has_alpha = 1;
            break;
        }
    if(has_alpha){
        while(len-->0){
            /* The joys of having a BGR byteorder */
            int alpha = src[3];
            *dst++ = (alpha*src[2]/255)<<16|
                     (alpha*src[1]/255)<< 8|
                     (alpha*src[0]/255)<< 0|
                     alpha<<24;
            src += 4;
        }
        return;
    }else{
        while(len-->0){
            *dst++ = src[3]<<16|
                     src[2]<< 8|
                     src[1]<< 0|
                     255<<24;
        }
    }
#else
    while(len-->0){
        *dst++ = src[3]<<16|
                 src[2]<< 8|
                 src[1]<< 0;
        src += 4;
    }
#endif
}


static void
_copy_pixels_xrgb(uint32_t *dst, const uint8_t *src, size_t len,const uint32_t *cmap)
{
    DEBUG_p("ARGB");
    while(len-->0){
#ifdef SUPPORT_ALPHA
        int alpha = src[0];
        *dst++ = (alpha*src[1]/255)<<16|
                 (alpha*src[2]/255)<< 8|
                 (alpha*src[3]/255)<< 0|
                 alpha<<24;
        src += 4;
#else
        *dst++ = src[1]<<16|
                 src[2]<< 8|
                 src[3]<< 0;
        src += 4;
#endif
    }
}


static void
_copy_pixels_bgr(uint32_t *dst, const uint8_t *src, size_t len,const uint32_t *cmap)
{
    while(len-->0){
        /* The joys of having a BGR byteorder */
        *dst++ = src[2]<<16|
                 src[1]<< 8|
                 src[0]<< 0;
        src += 3;
    }
}


static void
_copy_pixels_rgb(uint32_t *dst, const uint8_t *src, size_t len,const uint32_t *cmap)
{
    while(len-->0){
        *dst++ = src[0]<<16|
                 src[1]<< 8|
                 src[2]<< 0;
        src += 3;
    }
}


static void
_copy_pixels_cmap(uint32_t *dst, const uint8_t *src, size_t len,const uint32_t *cmap)
{
    while(len-->0)
        *dst++ = cmap[*src++];
}


static void
_copy_pixels_bitonal(uint32_t *dst, const uint8_t *src, size_t len,const uint32_t *cmap)
{
    int x;
    for(x=0;len-->0;++x){
        int bits;
        if(!(x&7))
            bits = *src++;
        bits<<=1;
        *dst++ = cmap[bits&0x100?1:0];
    }
}


static int
_process_header(abydos_plugin_handle_t *h,header_t *header)
{
    cairo_format_t format = CAIRO_FORMAT_RGB24;

    h->state.header_type = be32toh(header->type);
    h->state.header_maptype = be32toh(header->maptype);
    if (h->state.header_type > RT_FORMAT_RGB || h->state.header_maptype > RMT_EQUAL_RGB) {
        h->info->error = "unsupported RAS image variation";
        return -1;
    }

    h->state.cmap_size = be32toh(header->maplength);
    if(h->state.cmap_size > 3*256){
        h->info->error = "RAS image has bogus header data";
        return -1;
    }

    h->info->width = be32toh(header->width);
    h->info->height = be32toh(header->height);

    /* 2019-04-11 was the day this became less trivial */
    switch(be32toh(header->depth)){
    case 1:
        h->state.line_bytes = h->info->width / 8;
        if(h->info->width & 7)
            ++h->state.line_bytes;
        h->state.copy_pixels = _copy_pixels_bitonal;
        if(0==h->state.cmap_size){
            /*
            h->state.cmap_size = 3*2;
            h->state.cmap = malloc(h->state.cmap_size);
            h->state.cmap[0] = 255;
            h->state.cmap[1] = 0;
            h->state.cmap[2] = 255;
            h->state.cmap[3] = 0;
            h->state.cmap[4] = 255;
            h->state.cmap[5] = 0;
            */
            h->cmap = malloc(2*sizeof(uint32_t));
            h->cmap[0] = 0xffffff;
            h->cmap[1] = 0;
        }
        break;
    case 8:
        h->state.line_bytes = h->info->width * 1;
        h->state.copy_pixels = _copy_pixels_cmap;
        break;
    case 24:
        h->state.line_bytes = h->info->width * 3;
        if(RT_FORMAT_RGB == h->state.header_type){
            h->state.copy_pixels = _copy_pixels_rgb;
        }else{
            h->state.copy_pixels = _copy_pixels_bgr;
        }
        break;
    case 32:
#ifdef SUPPORT_ALPHA
        format = CAIRO_FORMAT_ARGB32;
#endif
        h->state.line_bytes = h->info->width * 4;
        if(RT_FORMAT_RGB == h->state.header_type){
            h->state.copy_pixels = _copy_pixels_xrgb;
        }else{
            h->state.copy_pixels = _copy_pixels_bgrx;
        }
        break;
    default:
        h->info->error = "RAS image has unknown type";
        return -1;
    }
    /* Now pad the line to be a multiple of 16 bits */
    h->state.line_bytes += h->state.line_bytes&1;

    h->state.line_buf = malloc(h->state.line_bytes);
#if 0
    if (!State->LineBuf) {
        h->info->error = "Not enough memory to load RAS image";
        return -1;
    }
#endif

    h->surface = cairo_image_surface_create(format,h->info->width,h->info->height);
#if 0
    if(!State->pixbuf){
        h->info->error = "Not enough memory to load RAS image";
        return -1;
    }
#endif

    if(h->state.info_func){
        /* Notify the client that we are ready to go */
        h->state.info_func(h->state.userdata);
    }

    return 0;
}


static void
_process_cmap(abydos_plugin_handle_t *h,const uint8_t *cmap)
{
    int colors = h->state.cmap_size/3;
    int i;
    h->cmap = (uint32_t*)malloc(256*sizeof(uint32_t));
    for(i=0;i<colors;++i){
        h->cmap[i] = cmap[i+colors*0] << 16|
                     cmap[i+colors*1] <<  8|
                     cmap[i+colors*2] <<  0;
    }
}


static void
_process_line(abydos_plugin_handle_t *h)
{
    uint8_t *src;
    uint32_t *dst;
    int stride;

    h->state.LineDone = 0;
    if (h->state.Lines >= h->info->height)
        return;

    //DEBUG_p("copying line %d",h->state.Lines);

    src = h->state.line_buf;
    dst = (uint32_t*)cairo_image_surface_get_data(h->surface);
    stride = cairo_image_surface_get_stride(h->surface)/sizeof(uint32_t);
    dst += h->state.Lines * stride;
    h->state.copy_pixels(dst,src,h->info->width,h->cmap);
    cairo_surface_mark_dirty(h->surface);

    if(h->state.progress_func){
        cairo_rectangle_int_t r;
        r.x = 0;
        r.y = h->state.Lines;
        r.width = h->info->width;
        r.height = 1;
        h->state.progress_func(h->state.userdata,&r);
    }

    ++h->state.Lines;
}


static void
_decompress(abydos_plugin_handle_t *h,abydos_buffer_t *buf)
{
    while(abydos_buffer_bytes(buf)>0){
        int b = abydos_buffer_pop(buf);
        switch (h->state.DecoderState) {
        case 0:
            if(0x80==b){
                h->state.DecoderState = 1;
            }else{
                h->state.line_buf[h->state.LineDone++] = b;
            }
            break;
        case 1:
            if(0==b){
                h->state.line_buf[h->state.LineDone++] = 0x80;
                h->state.DecoderState = 0;
            }else{
                h->state.DecoderState = b+1;
            }
            break;
        default:
            for (; h->state.DecoderState; --h->state.DecoderState) {
                h->state.line_buf[h->state.LineDone++] = b;
                if(h->state.LineDone >= h->state.line_bytes && h->state.line_bytes > 0)
                    _process_line(h);
            }
            break;
        }
        if(h->state.LineDone >= h->state.line_bytes && h->state.line_bytes > 0)
            _process_line(h);
    }
}


static int
_sunras_progressive_feed(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    abydos_buffer_add(&h->buf, (const unsigned char*)data, len);
    switch(h->parse_state){
    case PS_HEADER:
        if(abydos_buffer_size(&h->buf) < HEADER_SIZE)
            break;
        _process_header(h,(header_t*)abydos_buffer_get(&h->buf,HEADER_SIZE));
        h->parse_state = PS_CMAP;
        /* fall through */
    case PS_CMAP:
        if(h->state.cmap_size){
            if(abydos_buffer_bytes(&h->buf) < h->state.cmap_size)
                break;
            _process_cmap(h,abydos_buffer_get(&h->buf,h->state.cmap_size));
        }
        h->parse_state = PS_DATA;
        /* fall through */
    case PS_DATA:
        if(RT_BYTE_ENCODED == h->state.header_type){
            _decompress(h,&h->buf);
        }else{
            while(abydos_buffer_bytes(&h->buf) >= h->state.line_bytes){
                abydos_buffer_copy(&h->buf,h->state.line_buf,h->state.line_bytes);
                _process_line(h);
            }
        }
        break;
    case PS_IGNORE:
        DEBUG_p("ignoring %d bytes",(int)abydos_buffer_size(&h->buf));
        abydos_buffer_clear(&h->buf);
        break;
    }
    abydos_buffer_privatize(&h->buf);
    return 0;
}


static int
_sunras_progressive_end(abydos_plugin_handle_t *h)
{
    /* FIXME this thing needs to report errors if
     * we have unused image data
     */

    if(h->cmap){
        free(h->cmap);
        h->cmap = NULL;
    }
    free(h->state.line_buf);
    h->state.line_buf = NULL;

    return 0;
}


static cairo_surface_t *
_sunras_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _sunras_get_mime_types,
        .new = _sunras_new,
        .free = _sunras_free,
        .progressive_begin = _sunras_progressive_begin,
        .progressive_feed = _sunras_progressive_feed,
        .progressive_end = _sunras_progressive_end,
        .get_image_surface = _sunras_get_image_surface
    };
    return &p;
}
