/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "pc88.h"

/*

82      f3      16      00      dd      2a      12      bc
2a      14      bc      dd      7e      00      32      1a
bc      47      3a      18      bc      cd      0c      bd
23      3e      80      bd      20      2b      3e      fe
bc      20      26      3a      17      bc

3c      32      17      bc      fe      5f      28      25
21      00      c0      dd      23      14      c2      2d
bc      3e      03      c3      3a      bc      f3      16
00      dd      2a      12      bc      2a      14      bc
c3      2d      bc

*/


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
_dv_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-davinci",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
_dv_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->info->threadsafe = 1;
    h->surface = NULL;
    return h;
}


static void
_dv_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static int
_dv_create_from_data(abydos_plugin_handle_t *h,const char *data, size_t len)
{
    uint8_t bitplane_bytes[32000*3];
    uint8_t *bitplane[3];
    pc88_cairo_color_t palette[8];
    const uint8_t *p;
    const uint8_t *end;
    int offs;
    int c = 0;

    p = (uint8_t*)data;
    end = p + len;

    bitplane[0] = &bitplane_bytes[32000*0];
    bitplane[1] = &bitplane_bytes[32000*1];
    bitplane[2] = &bitplane_bytes[32000*2];
    for(offs=0;offs<16000;++offs){
        int b[3];
        if(c <= 0){
            if(p+4 > end)
                goto error_truncated;
            c = *p++;
            if(c & 128){
                /* It's a runlength. */
                c -= 128;
                b[0] = *p++;
                b[1] = *p++;
                b[2] = *p++;
            }else{
                b[0] = -1;
                if(p + c*3 > end)
                    goto error_truncated;
            }
        }
        if(b[0] >= 0){
            bitplane[0][offs] = b[0];
            bitplane[1][offs] = b[1];
            bitplane[2][offs] = b[2];
        }else{
            bitplane[0][offs] = *p++;
            bitplane[1][offs] = *p++;
            bitplane[2][offs] = *p++;
        }
        --c;
    }

    /* Set the default palette. */
    pc88_default_palette(palette);
    h->surface = pc88_decode_image(bitplane, palette);
    //h->surface = abydos_image_surface16_from_bitplanes(bitplane, 640, 200, 80, 3, palette, 0);

    h->info->width = 640;
    h->info->height = 200;
    h->info->pixel_ratio = 5.0/12;

    return 0;
error_truncated:
    DEBUG_p("truncated data");
    h->info->error = "Truncated data";
    return -1;
}


static cairo_surface_t *
_dv_get_image_surface(abydos_plugin_handle_t *h,
                       int page,
                       cairo_bool_t *layer_visibility,
                       int variant,
                       double time,
                       int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = _dv_get_mime_types,
        .new = _dv_new,
        .free = _dv_free,
        .create_from_data = _dv_create_from_data,
        .get_image_surface = _dv_get_image_surface
    };
    return &p;
}
