/*
 * Copyright (C) 2019 - 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <assert.h>
#include <pixman.h>
#include <string.h>                     /* memcpy() */

#include "pixels.h"

/* #define ENABLE_PIXMAN */


#ifdef ENABLE_PIXMAN
static void
_pixman_blit(uint32_t *src, uint32_t *dst, int src_stride, int dst_stride, int width, int height)
{
    pixman_blt(src, dst, src_stride, dst_stride, 8, 8, 0, 0, 0, 0, width, height);
}


static void
_pixman_transfer(pixman_format_code_t src_format, pixman_format_code_t dst_format, uint32_t *src, uint32_t *dst, int src_stride, int dst_stride, int width, int height)
{
    pixman_image_t *src_image = pixman_image_create_bits(src_format, width, height, src, src_stride);
    pixman_image_t *dst_image = pixman_image_create_bits(dst_format, width, height, dst, dst_stride);
    pixman_image_composite32(PIXMAN_OP_SRC, src_image, NULL, dst_image, 0, 0, 0, 0, 0, 0, width, height);
    pixman_image_unref(src_image);
    pixman_image_unref(src_image);
}
#endif /* ENABLE_PIXMAN */


cairo_surface_t *abydos_image_surface_create_from_pixels(abydos_pixel_format_t format,int flags,const uint8_t *src,int width,int height,int stride,int channels,uint32_t *palette)
{
    cairo_surface_t *surface;
    uint32_t *dst;
    int dst_skip;
    int src_skip;
    if(ABYDOS_PIXEL_FORMAT_UNKNOWN==format){
        /* guess pixelformat */
        switch(channels){
        case 1:
            if(palette){
                format = ABYDOS_PIXEL_FORMAT_c8;
            }else{
                format = ABYDOS_PIXEL_FORMAT_g8;
            }
            break;
        case 3:
            format = ABYDOS_PIXEL_FORMAT_r8g8b8;
            break;
        case 4:
            format = ABYDOS_PIXEL_FORMAT_r8g8b8a8;
            break;
        default:
            return NULL;
        }
    }
    surface = cairo_image_surface_create(has_alpha(format) ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24,width,height);
    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t)-width;
    src_skip = stride>0 ? stride-width*channels : 0;
    if(flags & ABYDOS_PIXEL_FLIPPED){
        src += stride * (height-1);
        src_skip -= stride*2;
    }
    switch(format){
    case ABYDOS_PIXEL_FORMAT_UNKNOWN:
        /* not reached */
        break;
    case ABYDOS_PIXEL_FORMAT_c8:
    case ABYDOS_PIXEL_FORMAT_c8a:
        assert(palette);
        while(height-->0){
            int x;
            for(x=width;x>0;--x)
                *dst++ = palette[*src++];
            src += src_skip;
            dst += dst_skip;
        }
        break;
    case ABYDOS_PIXEL_FORMAT_r8g8b8:
    case ABYDOS_PIXEL_FORMAT_x8r8g8b8:
    case ABYDOS_PIXEL_FORMAT_b8g8r8:
    case ABYDOS_PIXEL_FORMAT_x8b8g8r8:
    case ABYDOS_PIXEL_FORMAT_g8:
        /* no alpha channel */
        if(ABYDOS_PIXEL_FORMAT_CAIRO_RGB24==format){
#ifdef ENABLE_PIXMAN
            _pixman_blit((uint32_t*)src, dst, stride, cairo_image_surface_get_stride(surface), width, height);
#else
            if(src_skip || dst_skip){
                while(height-->0){
                    memcpy(dst,src,width*4);
                    src += src_skip;
                    dst += dst_skip;
                }
            }else{
                memcpy(dst,src,width*height*4);
            }
#endif
        }else if(is_gray(format)){
            while(height-->0){
                int x;
                for(x=width;x>0;--x){
                    int g = src[((channel_order_t)format).g];
                    *dst++ = g<<16|g<<8|g<<0;
                    src += channels;
                }
                src += src_skip;
                dst += dst_skip;
            }
        }else{
            while(height-->0){
                int x;
                for(x=width;x>0;--x){
                    int r = src[((channel_order_t)format).r];
                    int g = src[((channel_order_t)format).g];
                    int b = src[((channel_order_t)format).b];
                    *dst++ = r<<16|g<<8|b<<0;
                    src += channels;
                }
                src += src_skip;
                dst += dst_skip;
            }
        }
        break;
    case ABYDOS_PIXEL_FORMAT_r8g8b8a8:
    case ABYDOS_PIXEL_FORMAT_b8g8r8a8:
    case ABYDOS_PIXEL_FORMAT_a8r8g8b8:
    case ABYDOS_PIXEL_FORMAT_a8b8g8r8:
    case ABYDOS_PIXEL_FORMAT_g8a8:
    case ABYDOS_PIXEL_FORMAT_a8g8:
        if(flags & ABYDOS_PIXEL_PREMULTIPLIED){
            if(ABYDOS_PIXEL_FORMAT_CAIRO_ARGB32==format){
                if(src_skip || dst_skip){
                    while(height-->0){
                        memcpy(dst,src,width*4);
                        src += src_skip;
                        dst += dst_skip;
                    }
                }else{
                    memcpy(dst,src,width*height*4);
                }
            }else if(is_gray(format)){
                while(height-->0){
                    int x;
                    for(x=width;x>0;--x){
                        int a = src[((channel_order_t)format).a];
                        int g = src[((channel_order_t)format).g];
                        *dst++ = g<<16|g<<8|g<<0|a<<24;
                        src += channels;
                    }
                    src += src_skip;
                    dst += dst_skip;
                }
            }else{
                while(height-->0){
                    int x;
                    for(x=width;x>0;--x){
                        int r = src[((channel_order_t)format).r];
                        int g = src[((channel_order_t)format).g];
                        int b = src[((channel_order_t)format).b];
                        int a = src[((channel_order_t)format).a];
                        *dst++ = r<<16|g<<8|b<<0|a<<24;
                        src += channels;
                    }
                    src += src_skip;
                    dst += dst_skip;
                }
            }
        }else{
            /* alpha channel (prescaling is needed) */
            if(is_gray(format)){
                while(height-->0){
                    int x;
                    for(x=width;x>0;--x){
                        int a = src[((channel_order_t)format).a];
                        int g = a*src[((channel_order_t)format).g]/255;
                        *dst++ = g<<16|g<<8|g<<0|a<<24;
                        src += channels;
                    }
                    src += src_skip;
                    dst += dst_skip;
                }
            }else{
                while(height-->0){
                    int x;
                    for(x=width;x>0;--x){
                        int a = src[((channel_order_t)format).a];
                        int r = a*src[((channel_order_t)format).r]/255;
                        int g = a*src[((channel_order_t)format).g]/255;
                        int b = a*src[((channel_order_t)format).b]/255;
                        *dst++ = r<<16|g<<8|b<<0|a<<24;
                        src += channels;
                    }
                    src += src_skip;
                    dst += dst_skip;
                }
            }
        }
        break;
    case ABYDOS_PIXEL_FORMAT_r16g16b16:
    case ABYDOS_PIXEL_FORMAT_r16g16b16a16:
    case ABYDOS_PIXEL_FORMAT_b16g16r16a16:
        /* FIXME!! */
        break;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}
