/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef vector_h
#define vector_h

#include <stddef.h>                     /* size_t */

typedef struct{
    int esize;
    void *data;
    size_t allocation;
    int members;
}array_t;

array_t *array_init(array_t*,int esize);
void array_done(array_t*);
void *array_append(array_t*,int count);
void *array_insert(array_t*,int pos,int count);
void array_delete(array_t*,int pos,int count);
#define array_length(a) ((a)->members)
#define array_bytes(a) ((a)->members*(a)->esize)
#define array_get(a,i) ((a)->data+((i)*(a)->esize))
#define array_index(a,p) ((((void*)p)-(a)->data)/(a)->esize)
void *array_copy(array_t*);


typedef struct{
    double x;
    double y;
}point_t;

typedef struct _shape_t shape_t;
typedef struct _keyframe_t keyframe_t;
typedef struct _anim_t anim_t;

typedef enum{
    VECTOR_GOTO,
    VECTOR_LINE,
    VECTOR_CURVE,
    VECTOR_CLOSE
}vector_segment_type_t;

typedef struct{
    vector_segment_type_t type;
    int point;
}vector_segment_t;

typedef struct{
    int segment;
    int segment_count;
    int point_count;
}vector_path_t;

typedef enum{
    VECTOR_FILLED = 0,
    VECTOR_OUTLINE = 1,
    VECTOR_JUSTLINE = 2,
}vector_shape_style_t;

typedef struct{
    array_t point_array;
    array_t segment_array;
    array_t shape_array;
    array_t anim_array;
    array_t keyframe_array;
}vector_buffer_t;

vector_buffer_t *vector_buffer_init(vector_buffer_t *vb, int shape_size, int keyframe_size, int anim_size);
void vector_buffer_done(vector_buffer_t *vb);

#define vector_buffer_get_point(vb, i) ((point_t*)array_get(&(vb)->point_array,i))
#define vector_buffer_point_index(vb, p) array_index(&(vb)->point_array,p)
#define vector_buffer_point_count(vb) array_length(&(vb)->point_array)
#define vector_buffer_add_points(vb,c) (point_t*)array_append(&(vb)->point_array,c)
#define vector_buffer_insert_points(vb, i, c) (point_t*)array_insert(&(vb)->point_array,i,c)
#define vector_buffer_delete_points(vb, i, c) array_delete(&(vb)->point_array,i,c)
#define vector_buffer_copy_points(vb) (point_t*)array_copy(&(vb)->point_array)

#define vector_buffer_get_segment(vb, i) ((vector_segment_t*)array_get(&(vb)->segment_array,i))
#define vector_buffer_segment_count(vb) array_length(&(vb)->segment_array)
#define vector_buffer_add_segment(vb) (vector_segment_t*)array_append(&(vb)->segment_array,1)
#define vector_buffer_insert_segment(vb, i) (vector_segment_t*)array_insert(&(vb)->segment_array,i,1)
#define vector_buffer_delete_segments(vb, i, c) array_delete(&(vb)->segment_array,i,c)
point_t *vector_buffer_push_segment(vector_buffer_t *vb, vector_path_t *path, vector_segment_type_t type);
#define vector_buffer_copy_segments(vb) (vector_segment_t*)array_copy(&(vb)->segment_array)

#define vector_buffer_get_shape(vb, i) ((shape_t*)array_get(&(vb)->shape_array,i))
#define vector_buffer_shape_index(vb, p) array_index(&(vb)->shape_array,p)
#define vector_buffer_shape_count(vb) array_length(&(vb)->shape_array)
#define vector_buffer_add_shape(vb) (shape_t*)array_append(&(vb)->shape_array,1)
#define vector_buffer_delete_shape(vb, i) array_delete(&(vb)->shape_array,i,1)
#define vector_buffer_copy_shapes(vb) (shape_t*)array_copy(&(vb)->shape_array)

#define vector_buffer_get_anim(vb, i) ((anim_t*)array_get(&(vb)->anim_array,i))
#define vector_buffer_anim_count(vb) array_length(&(vb)->anim_array)
#define vector_buffer_add_anim(vb) (anim_t*)array_append(&(vb)->anim_array,1)
#define vector_buffer_copy_anims(vb) (anim_t*)array_copy(&(vb)->anim_array)

#define vector_buffer_keyframe_count(vb) array_length(&(vb)->keyframe_array)
#define vector_buffer_begin_keyframe(vb) (keyframe_t*)array_append(&(vb)->keyframe_array,1)
void vector_buffer_end_keyframe(vector_buffer_t*);
#define vector_buffer_copy_keyframes(vb) (keyframe_t*)array_copy(&(vb)->keyframe_array)

#endif /* vector_h */
